@@ -21,7 +21,7 @@
 
 #include <R.h>
 #include <Rinternals.h>
-#include "git2.h"
+#include <git2.h>
 
 SEXP git2r_signature_default(SEXP repo);
 int git2r_signature_from_arg(git_signature **out, SEXP signature);
